using System;
using System.IO;
using System.Threading;
using System.Net;
using System.Net.Sockets;
using System.Configuration;
using System.Reflection;
using System.Diagnostics;
using Docsultant.Samples.Net.Common;

namespace Docsultant.Samples.Net.Redirector
{
	/// <summary>
	/// TcpHack - extract Socket object from TcpClient one
	/// </summary>
	class TcpHack
	{
		public static Socket GetSocket(TcpClient client) 
		{
			PropertyInfo pi = client.GetType().GetProperty("Client", BindingFlags.NonPublic | BindingFlags.Instance);
			return (Socket)pi.GetValue(client, null);
		}
	};

	/// <summary>
	/// TcpUnaryLink - thread that manages unidirectional socket link. 
	/// Reads data from one socket and forwards to another one.
	/// </summary>
	class TcpUnaryLink
	{
		private NetworkStream m_stmIn;
		private NetworkStream m_stmOut;
		private String        m_sHeader;

		public TcpUnaryLink(TcpClient tcpIn, TcpClient tcpOut)
		{
			m_stmIn  = tcpIn.GetStream();
			m_stmOut = tcpOut.GetStream();
			m_sHeader = TcpHack.GetSocket(tcpIn).RemoteEndPoint + " -> " + TcpHack.GetSocket(tcpOut).RemoteEndPoint;
			Thread t = new Thread(new ThreadStart(this.ThreadProc));
			t.Start();
		}
		
		public void ThreadProc()
		{
			Debug.WriteLine("Enter TcpUnaryLink::ThreadProc");
			int cbBuf = 4096;
			byte[] aBuf = new byte[cbBuf];
			int cbRead = 0;

			try
			{
				while(true)
				{
					cbRead = m_stmIn.Read(aBuf, 0, cbBuf);
					if( cbRead>0 )
					{
						Trace.WriteLine("send "+cbRead+" raw bytes:", m_sHeader);
						Logger.Log(aBuf, 0, cbRead);
						m_stmOut.Write(aBuf, 0, cbRead);
					} else
						break;
				}
			} catch (Exception e)
			{
				Debug.WriteLine(e);
			}

			try
			{ 
				m_stmIn.Close(); 
				m_stmOut.Close(); 
			} 
			finally
			{
				Trace.WriteLine("close connection", m_sHeader);
			}

			Debug.WriteLine("Leave TcpUnaryLink::ThreadProc");
		}
	};

	/// <summary>
	/// TcpLinkFactory - creates TcpUnariLink connection in separate thread.
	/// Provides asynchronous functionality.
	/// </summary>
	class TcpLinkFactory
	{
		private String    m_sHost;
		private int       m_nPort;
		private TcpClient m_tcpIn;

		private TcpLinkFactory(String sHost, int nPort, TcpClient tcpIn)
		{
			m_sHost = sHost;
			m_nPort = nPort;
			m_tcpIn = tcpIn;
		}

		private void ThreadProc()
		{
			try
			{
				TcpClient tcpOut = new TcpClient(m_sHost, m_nPort);
				TcpUnaryLink link1 = new TcpUnaryLink(m_tcpIn, tcpOut);
				TcpUnaryLink link2 = new TcpUnaryLink(tcpOut, m_tcpIn);
			} catch (Exception e)
			{
				Debug.WriteLine(e);
				Trace.WriteLine("failed connect to "+m_sHost+":"+m_nPort+", connection closed", TcpHack.GetSocket(m_tcpIn).RemoteEndPoint.ToString());
				m_tcpIn.Close();
			}
		}

		public static void Create(String sHost, int nPort, TcpClient tcpIn)
		{
			TcpLinkFactory f = new TcpLinkFactory(sHost, nPort, tcpIn);
			Thread t = new Thread(new ThreadStart(f.ThreadProc));
			t.Start();
		}
	}


	/// <summary>
	/// RedirectorListener - listener
	/// </summary>
	class RedirectorListener : TcpListenerThread
	{
		public RedirectorListener():base(Redirector.LocalHost, Redirector.LocalPort)
		{
		}

		public override bool OnAccept(TcpClient tcp)		
		{
			TcpLinkFactory.Create(Redirector.RemoteHost, Redirector.RemotePort, tcp);
			Trace.WriteLine("Accept incoming connection: "+ TcpHack.GetSocket(tcp).RemoteEndPoint);
			return true;
		}
	}
	
	/// <summary>
	/// Redirector - proxy application that accepts incoming connections,
	/// forwards it to the specified host, and logs all trafic to text file.
	/// </summary>
	class Redirector : AppConfig
	{
		private static int m_nLocalPort = -1;
		public static int LocalPort
		{
			get{ return m_nLocalPort<0? (m_nLocalPort = getInt("LocalPort", 7000)):m_nLocalPort; }
		}

		private static String m_sLocalHost;
		public static String LocalHost
		{
			get{ return m_sLocalHost==null? (m_sLocalHost=getStr("LocalHost", "localhost")):m_sLocalHost; }
		}

		private static int m_nRemotePort = -1;
		public static int RemotePort
		{
			get{ return m_nRemotePort<0? (m_nRemotePort = getInt("RemotePort", 5000)):m_nRemotePort ; }
		}

		private static String m_sRemoteHost; 
		public static String RemoteHost
		{
			get{ return m_sRemoteHost==null? (m_sRemoteHost=getStr("RemoteHost", "localhost")):m_sRemoteHost; }
		}

		private void Run()
		{
			Trace.WriteLine("----------------------------------------");
			Trace.WriteLine("Starting redirector...");
			Trace.WriteLine(" listening  on \""+LocalHost+"\", port "+LocalPort);
			Trace.WriteLine(" forwarding to \""+RemoteHost+"\", port "+RemotePort);
			RedirectorListener rl = new RedirectorListener();
			rl.Start();
		}

		public static void Main(string[] args)
		{
			if( args.Length>=1 )
			{
				if( args[0]=="?" || args[0]=="--help" || args[0]=="-h" )
				{
					Console.WriteLine("TCP redirector utility.");
					Console.WriteLine("Copyright 2002 (c) by Vadim Melnik, All right reserved.");
					Console.WriteLine("");
					Console.WriteLine("Usage: redirector.exe [options, ...] ");
					Console.WriteLine("Command line options:");
					Console.WriteLine("  -h, ?, --help     : this help");
					Console.WriteLine("  -host    <addr>   : specifies remote host address");
					Console.WriteLine("  -port    <port>   : specifies remote host port");
					Console.WriteLine("  -lhost   <port>   : specifies local host address");
					Console.WriteLine("  -lport   <port>   : specifies local port");
					Console.WriteLine("  -logfile <file>   : specifies log file name");
					Console.WriteLine("");
					return;
				}

				int i1 = 0, i2 = 1;
				while( i2<args.Length )
				{
					string s1 = args[i1];
					i1+=2; 
					
					string s2 = args[i2];	
					i2+=2;

					switch( s1.ToLower() )
					{
						case "-host":
								m_sRemoteHost = s2;
							break;

						case "-port":
								m_nRemotePort = Int32.Parse(s2);
							break;
						
						case "-lhost":
								m_sLocalHost = s2;
							break;

						case "-lport":
								m_nLocalPort = Int32.Parse(s2);
							break;

						case "-logfile":
								Logger.LogFile = s2;
							break;

						default:
								i1--; i2--;
							break;
					}
				}
			}

			Logger.Init();
			Redirector r = new Redirector();
			r.Run();
		}

	}
}
